"""
Useful utilities for working with raster data, typically `numpy` arrays.

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Domagoj Korais, Matic Lubej, Žiga Lukšič (Sinergise)
Copyright (c) 2017-2022 Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
from typing import Tuple

import numpy as np

from .types import Literal


def fast_nanpercentile(data: np.ndarray, percentile: float, *, method: str = "linear") -> np.ndarray:
    """This is an alternative implementation of `numpy.nanpercentile`. For cases where the size of the first dimension
    is relatively small compared to the size of the entire array it works much faster than the original.

    The algorithm divides pixel data over the first axis into groups by how many NaN values they have. In each group
    NaN values are removed and `numpy.percentile` function is applied. If any series contains only NaN values also any
    percentile of that series will be NaN.

    This function differs from `numpy` implementations only in the following:

    - In case the size of the first dimension of `data` is `0` this method will still return an output array with
      all NaN values. This matches with `numpy.nanpercentile` while `numpy.percentile` raises an error.
    - The output dtype of this method will be always the same as the input dtype while `numpy` implementations
      in many cases use `float64` as the output dtype.

    :param data: An array for which percentiles will be calculated along the first axis.
    :param percentile: A percentile to compute, which must be between `0` and `100` inclusive.
    :param method: A method for estimating the percentile. This parameter is propagated to `numpy.percentile`.
    :return: An array of percentiles and a shape equal to the shape of `data` array without the first dimension.
    """
    method_kwargs = {"method" if np.__version__ >= "1.22.0" else "interpolation": method}

    combined_data = np.zeros(data.shape[1:], dtype=data.dtype)

    no_data_counts = np.count_nonzero(np.isnan(data), axis=0)
    for no_data_num in np.unique(no_data_counts):
        mask = no_data_counts == no_data_num

        chunk = data[..., mask]
        time_size, sample_size = chunk.shape

        if time_size == no_data_num:
            result = np.full(sample_size, np.nan, dtype=chunk.dtype)
        else:
            chunk = chunk.flatten(order="F")
            chunk = chunk[~np.isnan(chunk)]
            chunk = chunk.reshape((time_size - no_data_num, sample_size), order="F")

            result = np.percentile(chunk, q=percentile, axis=0, **method_kwargs)  # type: ignore[call-overload]

        combined_data[mask] = result

    return combined_data


def constant_pad(
    array: np.ndarray,
    multiple_of: Tuple[int, int],
    up_down_rule: Literal["even", "up", "down"] = "even",
    left_right_rule: Literal["even", "left", "right"] = "even",
    pad_value: float = 0,
) -> np.ndarray:
    """Function pads an image of shape (rows, columns, channels) with zeros.

    It pads an image so that the shape becomes (rows + padded_rows, columns + padded_columns, channels), where
    padded_rows = (int(rows/multiple_of[0]) + 1) * multiple_of[0] - rows

    Same rule is applied to columns.

    :param array: Array with shape `(rows, columns, ...)` to be padded.
    :param multiple_of: make array' rows and columns multiple of this tuple
    :param up_down_rule: Add padded rows evenly to the top/bottom of the image, or up (top) / down (bottom) only
    :param left_right_rule: Add padded columns evenly to the left/right of the image, or left / right only
    :param pad_value: Value to be assigned to padded rows and columns
    """
    rows, columns = array.shape[:2]
    row_padding, col_padding = 0, 0

    if rows % multiple_of[0]:
        row_padding = (int(rows / multiple_of[0]) + 1) * multiple_of[0] - rows

    if columns % multiple_of[1]:
        col_padding = (int(columns / multiple_of[1]) + 1) * multiple_of[1] - columns

    row_padding_up, row_padding_down, col_padding_left, col_padding_right = 0, 0, 0, 0

    if row_padding > 0:
        if up_down_rule == "up":
            row_padding_up = row_padding
        elif up_down_rule == "down":
            row_padding_down = row_padding
        elif up_down_rule == "even":
            row_padding_up = int(row_padding / 2)
            row_padding_down = row_padding_up + (row_padding % 2)
        else:
            raise ValueError("Padding rule for rows not supported. Choose between even, down or up!")

    if col_padding > 0:
        if left_right_rule == "left":
            col_padding_left = col_padding
        elif left_right_rule == "right":
            col_padding_right = col_padding
        elif left_right_rule == "even":
            col_padding_left = int(col_padding / 2)
            col_padding_right = col_padding_left + (col_padding % 2)
        else:
            raise ValueError("Padding rule for columns not supported. Choose between even, left or right!")

    return np.lib.pad(
        array,
        ((row_padding_up, row_padding_down), (col_padding_left, col_padding_right)),
        "constant",
        constant_values=((pad_value, pad_value), (pad_value, pad_value)),
    )



"""
A module implementing EOPatch merging utility

Credits:
Copyright (c) 2018-2020 William Ouellette
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
from __future__ import annotations

import datetime as dt
import functools
import itertools as it
import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from sentinelhub import BBox

from .constants import FeatureType
from .exceptions import EORuntimeWarning
from .utils.parsing import FeatureParser, FeatureSpec, FeaturesSpecification
from .utils.types import Literal

if TYPE_CHECKING:
    from .eodata import EOPatch

OperationInputType = Union[Literal[None, "concatenate", "min", "max", "mean", "median"], Callable]


def merge_eopatches(
    *eopatches: EOPatch,
    features: FeaturesSpecification = ...,
    time_dependent_op: OperationInputType = None,
    timeless_op: OperationInputType = None,
) -> Dict[FeatureSpec, Any]:
    """Merge features of given EOPatches into a new EOPatch.

    :param eopatches: Any number of EOPatches to be merged together
    :param features: A collection of features to be merged together. By default, all features will be merged.
    :param time_dependent_op: An operation to be used to join data for any time-dependent raster feature. Before
        joining time slices of all arrays will be sorted. Supported options are:

        - None (default): If time slices with matching timestamps have the same values, take one. Raise an error
          otherwise.
        - 'concatenate': Keep all time slices, even the ones with matching timestamps
        - 'min': Join time slices with matching timestamps by taking minimum values. Ignore NaN values.
        - 'max': Join time slices with matching timestamps by taking maximum values. Ignore NaN values.
        - 'mean': Join time slices with matching timestamps by taking mean values. Ignore NaN values.
        - 'median': Join time slices with matching timestamps by taking median values. Ignore NaN values.

    :param timeless_op: An operation to be used to join data for any timeless raster feature. Supported options
        are:

        - None (default): If arrays are the same, take one. Raise an error otherwise.
        - 'concatenate': Join arrays over the last (i.e. bands) dimension
        - 'min': Join arrays by taking minimum values. Ignore NaN values.
        - 'max': Join arrays by taking maximum values. Ignore NaN values.
        - 'mean': Join arrays by taking mean values. Ignore NaN values.
        - 'median': Join arrays by taking median values. Ignore NaN values.

    :return: Merged EOPatch
    """
    reduce_timestamps = time_dependent_op != "concatenate"
    time_dependent_operation = _parse_operation(time_dependent_op, is_timeless=False)
    timeless_operation = _parse_operation(timeless_op, is_timeless=True)

    feature_parser = FeatureParser(features)
    all_features = {feature for eopatch in eopatches for feature in feature_parser.get_features(eopatch)}
    eopatch_content: Dict[FeatureSpec, object] = {}

    timestamps, order_mask_per_eopatch = _merge_timestamps(eopatches, reduce_timestamps)
    optimize_raster_temporal = _check_if_optimize(eopatches, time_dependent_op)

    for feature in all_features:
        feature_type, feature_name = feature

        if feature_type.is_raster():
            if feature_type.is_temporal():
                eopatch_content[feature] = _merge_time_dependent_raster_feature(
                    eopatches, feature, time_dependent_operation, order_mask_per_eopatch, optimize_raster_temporal
                )
            else:
                eopatch_content[feature] = _merge_timeless_raster_feature(eopatches, feature, timeless_operation)

        if feature_type.is_vector():
            eopatch_content[feature] = _merge_vector_feature(eopatches, feature)

        if feature_type is FeatureType.TIMESTAMP:
            eopatch_content[feature] = timestamps

        if feature_type is FeatureType.META_INFO:
            feature_name = cast(str, feature_name)  # parser makes sure of it
            eopatch_content[feature] = _select_meta_info_feature(eopatches, feature_name)

        if feature_type is FeatureType.BBOX:
            eopatch_content[feature] = _get_common_bbox(eopatches)

    return eopatch_content


def _parse_operation(operation_input: OperationInputType, is_timeless: bool) -> Callable:
    """Transforms operation's instruction (i.e. an input string) into a function that can be applied to a list of
    arrays. If the input already is a function it returns it.
    """
    defaults: Dict[Optional[str], Callable] = {
        None: _return_if_equal_operation,
        "concatenate": functools.partial(np.concatenate, axis=-1 if is_timeless else 0),
        "mean": functools.partial(np.nanmean, axis=0),
        "median": functools.partial(np.nanmedian, axis=0),
        "min": functools.partial(np.nanmin, axis=0),
        "max": functools.partial(np.nanmax, axis=0),
    }
    if operation_input in defaults:
        return defaults[operation_input]  # type: ignore[index]

    if isinstance(operation_input, Callable):  # type: ignore[arg-type]  #mypy 0.981 has issues with callable
        return cast(Callable, operation_input)
    raise ValueError(f"Merge operation {operation_input} is not supported")


def _return_if_equal_operation(arrays: np.ndarray) -> bool:
    """Checks if arrays are all equal and returns first one of them. If they are not equal it raises an error."""
    if _all_equal(arrays):
        return arrays[0]
    raise ValueError("Cannot merge given arrays because their values are not the same.")


def _merge_timestamps(
    eopatches: Sequence[EOPatch], reduce_timestamps: bool
) -> Tuple[List[dt.datetime], List[np.ndarray]]:
    """Merges together timestamps from EOPatches. It also prepares a list of masks, one for each EOPatch, how
    timestamps should be ordered and joined together.
    """
    timestamps_per_eopatch = [eopatch.timestamp for eopatch in eopatches]
    all_timestamps = [timestamp for eopatch_timestamps in timestamps_per_eopatch for timestamp in eopatch_timestamps]

    if not all_timestamps:
        return [], [np.array([], dtype=np.int32) for _ in range(len(eopatches))]

    if reduce_timestamps:
        unique_timestamps, order_mask = np.unique(all_timestamps, return_inverse=True)  # type: ignore[call-overload]
        ordered_timestamps = unique_timestamps.tolist()
    else:
        order_mask = np.argsort(all_timestamps)  # type: ignore[arg-type]
        ordered_timestamps = sorted(all_timestamps)

    order_mask = order_mask.tolist()

    order_mask_iter = iter(order_mask)
    order_mask_per_eopatch = [
        np.array(list(it.islice(order_mask_iter, len(eopatch_timestamps))), dtype=np.int32)
        for eopatch_timestamps in timestamps_per_eopatch
    ]

    return ordered_timestamps, order_mask_per_eopatch


def _check_if_optimize(eopatches: Sequence[EOPatch], operation_input: OperationInputType) -> bool:
    """Checks whether optimisation of `_merge_time_dependent_raster_feature` is possible"""
    if operation_input not in [None, "mean", "median", "min", "max"]:
        return False
    timestamp_list = [eopatch.timestamp for eopatch in eopatches]
    return _all_equal(timestamp_list)


def _merge_time_dependent_raster_feature(
    eopatches: Sequence[EOPatch],
    feature: FeatureSpec,
    operation: Callable,
    order_mask_per_eopatch: Sequence[np.ndarray],
    optimize: bool,
) -> np.ndarray:
    """Merges numpy arrays of a time-dependent raster feature with a given operation and masks on how to order and join
    time raster's time slices.
    """

    merged_array, merged_order_mask = _extract_and_join_time_dependent_feature_values(
        eopatches, feature, order_mask_per_eopatch, optimize
    )

    # Case where feature array is already in the correct order and doesn't need splitting, which includes a case
    # where array has a size 0
    if _is_strictly_increasing(merged_order_mask):
        return merged_array

    sort_mask = np.argsort(merged_order_mask)
    merged_array = merged_array[sort_mask]
    merged_order_mask = merged_order_mask[sort_mask]

    # Case where feature array has been sorted but doesn't need splitting
    if _is_strictly_increasing(merged_order_mask):
        return merged_array

    split_indices = np.nonzero(np.diff(merged_order_mask))[0] + 1
    split_arrays = np.split(merged_array, split_indices)
    del merged_array

    try:
        split_arrays = [operation(array_chunk) for array_chunk in split_arrays]
    except ValueError as exception:
        raise ValueError(
            f"Failed to merge {feature} with {operation}, try setting a different value for merging "
            "parameter time_dependent_op"
        ) from exception

    return np.array(split_arrays)


def _extract_and_join_time_dependent_feature_values(
    eopatches: Sequence[EOPatch],
    feature: FeatureSpec,
    order_mask_per_eopatch: Sequence[np.ndarray],
    optimize: bool,
) -> Tuple[np.ndarray, np.ndarray]:
    """Collects feature arrays from EOPatches that have them and joins them together. It also joins together
    corresponding order masks.
    """
    arrays = []
    order_masks = []
    feature_type, feature_name = feature

    for eopatch, order_mask in zip(eopatches, order_mask_per_eopatch):
        if feature_name in eopatch[feature_type]:
            array = eopatch[feature_type, feature_name]
            if order_mask.size != array.shape[0]:
                raise ValueError(
                    f"Cannot merge a time-dependent feature {feature} because time dimension of an array "
                    f"in one EOPatch is {array.shape[0]} but EOPatch has {order_mask.size} timestamps"
                )

            arrays.append(array)
            order_masks.append(order_mask)

    if len(arrays) == 1 or (optimize and _all_equal(arrays)):
        return arrays[0], order_masks[0]
    return np.concatenate(arrays, axis=0), np.concatenate(order_masks)


def _is_strictly_increasing(array: np.ndarray) -> bool:
    """Checks if a 1D array of values is strictly increasing."""
    return (np.diff(array) > 0).all().astype(bool)


def _merge_timeless_raster_feature(
    eopatches: Sequence[EOPatch], feature: FeatureSpec, operation: Callable
) -> np.ndarray:
    """Merges numpy arrays of a timeless raster feature with a given operation."""
    arrays = _extract_feature_values(eopatches, feature)

    if len(arrays) == 1:
        return arrays[0]

    try:
        return operation(arrays)
    except ValueError as exception:
        raise ValueError(
            f"Failed to merge {feature} with {operation}, try setting a different value for merging "
            "parameter timeless_op."
        ) from exception


def _merge_vector_feature(eopatches: Sequence[EOPatch], feature: FeatureSpec) -> GeoDataFrame:
    """Merges GeoDataFrames of a vector feature."""
    dataframes = _extract_feature_values(eopatches, feature)

    if len(dataframes) == 1:
        return dataframes[0]

    crs_list = [dataframe.crs for dataframe in dataframes if dataframe.crs is not None]
    if not crs_list:
        crs_list = [None]
    if not _all_equal(crs_list):
        raise ValueError(f"Cannot merge feature {feature} because dataframes are defined for different CRS")

    merged_dataframe = GeoDataFrame(pd.concat(dataframes, ignore_index=True), crs=crs_list[0])
    merged_dataframe = merged_dataframe.drop_duplicates(ignore_index=True)
    # In future a support for vector operations could be added here

    return merged_dataframe


def _select_meta_info_feature(eopatches: Sequence[EOPatch], feature_name: str) -> Any:
    """Selects a value for a meta info feature of a merged EOPatch. By default, the value is the first one."""
    values = _extract_feature_values(eopatches, (FeatureType.META_INFO, feature_name))

    if not _all_equal(values):
        message = (
            f"EOPatches have different values of meta info feature {feature_name}. The first value will be "
            "used in a merged EOPatch."
        )
        warnings.warn(message, category=EORuntimeWarning)

    return values[0]


def _get_common_bbox(eopatches: Sequence[EOPatch]) -> Optional[BBox]:
    """Makes sure that all EOPatches, which define a bounding box and CRS, define the same ones."""
    bboxes = [eopatch.bbox for eopatch in eopatches if eopatch.bbox is not None]

    if not bboxes:
        return None

    if _all_equal(bboxes):
        return bboxes[0]
    raise ValueError("Cannot merge EOPatches because they are defined for different bounding boxes.")


def _extract_feature_values(eopatches: Sequence[EOPatch], feature: FeatureSpec) -> List[Any]:
    """A helper function that extracts a feature values from those EOPatches where a feature exists."""
    feature_type, feature_name = feature
    return [eopatch[feature] for eopatch in eopatches if feature_name in eopatch[feature_type]]


def _all_equal(values: Union[Sequence[Any], np.ndarray]) -> bool:
    """A helper function that checks if all values in a given list are equal to each other."""
    first_value = values[0]

    if isinstance(first_value, np.ndarray):
        is_numeric_dtype = np.issubdtype(first_value.dtype, np.number)
        return all(np.array_equal(first_value, array, equal_nan=is_numeric_dtype) for array in values[1:])

    return all(first_value == value for value in values[1:])



"""
This module implements the core class hierarchy for implementing EO tasks. An EO task is any class the inherits
from the abstract EOTask class. Each EO task has to implement the execute method; invoking __call__ on a EO task
instance invokes the execute method. EO tasks are meant primarily to operate on EO patches (i.e. instances of EOPatch).

EO task classes are generally lightweight (i.e. not too complicated), short, and do one thing well. For example, an
EO task might take as input an EOPatch containing cloud mask and return as a result the cloud coverage for that mask.

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Nejc Vesel, Jovan Višnjić, Anže Zupanc (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import inspect
import logging
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Iterable, Type, TypeVar, Union

from .constants import FeatureType
from .utils.parsing import (
    FeatureParser,
    FeaturesSpecification,
    parse_feature,
    parse_features,
    parse_renamed_feature,
    parse_renamed_features,
)
from .utils.types import EllipsisType

LOGGER = logging.getLogger(__name__)

Self = TypeVar("Self")


class EOTask(metaclass=ABCMeta):
    """Base class for EOTask."""

    parse_feature = staticmethod(parse_feature)
    parse_renamed_feature = staticmethod(parse_renamed_feature)
    parse_features = staticmethod(parse_features)
    parse_renamed_features = staticmethod(parse_renamed_features)

    def __new__(cls: Type[Self], *args: Any, **kwargs: Any) -> Self:
        """Stores initialization parameters and the order to the instance attribute `init_args`."""
        self = super().__new__(cls)  # type: ignore[misc]

        init_args: Dict[str, object] = {}
        for arg, value in zip(inspect.getfullargspec(self.__init__).args[1 : len(args) + 1], args):
            init_args[arg] = repr(value)
        for arg in inspect.getfullargspec(self.__init__).args[len(args) + 1 :]:
            if arg in kwargs:
                init_args[arg] = repr(kwargs[arg])

        self._private_task_config = _PrivateTaskConfig(init_args=init_args)

        return self

    @property
    def private_task_config(self) -> "_PrivateTaskConfig":
        """Keeps track of the arguments for which the task was initialized for better logging.

        :return: The initial configuration arguments of the task
        """
        return self._private_task_config  # type: ignore[attr-defined]

    def __call__(self, *eopatches, **kwargs):
        """Syntactic sugar for task execution"""
        return self.execute(*eopatches, **kwargs)

    @abstractmethod
    def execute(self, *eopatches, **kwargs):
        """Override to specify action performed by task."""

    @staticmethod
    def get_feature_parser(
        features: FeaturesSpecification, allowed_feature_types: Union[Iterable[FeatureType], EllipsisType] = ...
    ) -> FeatureParser:
        """See :class:`FeatureParser<eolearn.core.utilities.FeatureParser>`."""
        return FeatureParser(features, allowed_feature_types=allowed_feature_types)


@dataclass(frozen=True)
class _PrivateTaskConfig:
    """A container for configuration parameters about an EOTask itself.

    :param init_args: A dictionary of parameters and values used for EOTask initialization
    """

    init_args: Dict[str, object]



"""
The utilities module is a collection of classes and functions used across the eolearn package, such as checking whether
two objects are deeply equal, padding of an image, etc.

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Tomislav Slijepčević, Nejc Vesel, Jovan Višnjić (Sinergise)
Copyright (c) 2017-2022 Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import uuid
from typing import Sequence, Union, cast

import geopandas as gpd
import numpy as np
from geopandas.testing import assert_geodataframe_equal


def deep_eq(fst_obj: object, snd_obj: object) -> bool:
    """Compares whether fst_obj and snd_obj are deeply equal.

    In case when both fst_obj and snd_obj are of type np.ndarray or either np.memmap, they are compared using
    np.array_equal(fst_obj, snd_obj). Otherwise, when they are lists or tuples, they are compared for length and then
    deep_eq is applied component-wise. When they are dict, they are compared for key set equality, and then deep_eq is
    applied value-wise. For all other data types that are not list, tuple, dict, or np.ndarray, the method falls back
    to the __eq__ method.

    Because np.ndarray is not a hashable object, it is impossible to form a set of numpy arrays, hence deep_eq works
    correctly.

    :param fst_obj: First object compared
    :param snd_obj: Second object compared
    :return: `True` if objects are deeply equal, `False` otherwise
    """
    # pylint: disable=too-many-return-statements
    if not isinstance(fst_obj, type(snd_obj)):
        return False

    if isinstance(fst_obj, np.ndarray):
        snd_obj = cast(np.ndarray, snd_obj)
        if fst_obj.dtype != snd_obj.dtype:
            return False
        fst_nan_mask = np.isnan(fst_obj)
        snd_nan_mask = np.isnan(snd_obj)
        return np.array_equal(fst_obj[~fst_nan_mask], snd_obj[~snd_nan_mask]) and np.array_equal(
            fst_nan_mask, snd_nan_mask
        )

    if isinstance(fst_obj, gpd.GeoDataFrame):
        try:
            # We allow differences in index types and in dtypes of columns
            assert_geodataframe_equal(fst_obj, snd_obj, check_index_type=False, check_dtype=False)
            return True
        except AssertionError:
            return False

    if isinstance(fst_obj, (tuple, list)):
        snd_obj = cast(Sequence, snd_obj)

        return len(fst_obj) == len(snd_obj) and all(map(deep_eq, fst_obj, snd_obj))

    if isinstance(fst_obj, dict):
        snd_obj = cast(dict, snd_obj)

        if fst_obj.keys() != snd_obj.keys():
            return False

        return all(deep_eq(fst_obj[key], snd_obj[key]) for key in fst_obj)

    return fst_obj == snd_obj


def generate_uid(prefix: str) -> str:
    """Generates a (sufficiently) unique ID starting with the `prefix`.

    The ID is composed of the prefix, a hexadecimal string obtained from the current time and a random hexadecimal
    string. This makes the uid sufficiently unique.
    """
    time_uid = uuid.uuid1(node=0).hex[:-12]
    random_uid = uuid.uuid4().hex[:12]
    return f"{prefix}-{time_uid}-{random_uid}"


def is_discrete_type(number_type: Union[np.dtype, type]) -> bool:
    """Checks if a given `numpy` type is a discrete numerical type."""
    return np.issubdtype(number_type, np.integer) or np.issubdtype(number_type, bool)



"""
Implementation of custom eo-learn exceptions and warnings

Credits:
Copyright (c) 2021-2022 Matej Aleksandrov, Žiga Lukšič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import warnings
from typing import Any, Type


class EODeprecationWarning(DeprecationWarning):
    """A custom deprecation warning for eo-learn package."""


class EOUserWarning(UserWarning):
    """A custom user warning for eo-learn package."""


class EORuntimeWarning(RuntimeWarning):
    """A custom runtime warning for eo-learn package."""


warnings.simplefilter("default", EODeprecationWarning)
warnings.simplefilter("default", EOUserWarning)
warnings.simplefilter("always", EORuntimeWarning)


def renamed_and_deprecated(deprecated_class: Type) -> Type:
    """A class decorator that signals that the class has been renamed when initialized.

    Example of use:

    .. code-block:: python

        @renamed_and_deprecated
        class OldNameForClass(NewNameForClass):
            ''' Deprecated version of `NewNameForClass`
            '''

    """

    def warn_and_init(self: Any, *args: Any, **kwargs: Any) -> None:
        warnings.warn(
            f"The class {self.__class__.__name__} has been renamed to {self.__class__.__mro__[1].__name__}. "
            "The old name is deprecated and will be removed in version 1.0",
            EODeprecationWarning,
        )
        super(deprecated_class, self).__init__(*args, **kwargs)

    deprecated_class.__init__ = warn_and_init
    return deprecated_class



"""
Types and type aliases used throughout the code.

Credits:
Copyright (c) 2022 Žiga Lukšič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
# pylint: disable=unused-import
import sys

if sys.version_info >= (3, 10):
    from types import EllipsisType  # pylint: disable=ungrouped-imports
else:
    import builtins  # noqa: F401

    from typing_extensions import TypeAlias

    EllipsisType: TypeAlias = "builtins.ellipsis"

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal  # pylint: disable=ungrouped-imports # noqa: F401



"""
Utilities for a basic Python parallelization that build on top of `concurrent.futures` module from Standard Python
library.

Credits:
Copyright (c) 2022 Matej Aleksandrov, Žiga Lukšič (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
import concurrent.futures
import multiprocessing
from concurrent.futures import FIRST_COMPLETED, Executor, Future, ProcessPoolExecutor, ThreadPoolExecutor
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Collection, Generator, Iterable, List, Optional, Tuple, TypeVar, cast

from tqdm.auto import tqdm

if TYPE_CHECKING:
    from threading import Lock

    MULTIPROCESSING_LOCK: Optional[Lock] = None
else:
    MULTIPROCESSING_LOCK = None

# pylint: disable=invalid-name
_T = TypeVar("_T")
_FutureType = TypeVar("_FutureType")
_OutputType = TypeVar("_OutputType")


class _ProcessingType(Enum):
    """Type of EOExecutor processing"""

    SINGLE_PROCESS = "single process"
    MULTIPROCESSING = "multiprocessing"
    MULTITHREADING = "multithreading"
    RAY = "ray"


def _decide_processing_type(workers: Optional[int], multiprocess: bool) -> _ProcessingType:
    """Decides processing type according to given parameters.

    :param workers: A number of workers to be used (either threads or processes). If a single worker is given it will
        always use the current thread and process.
    :param multiprocess: A flag to decide between multiple processes and multiple threads in a single process.
    :return: An enum defining a type of processing.
    """
    if workers == 1:
        return _ProcessingType.SINGLE_PROCESS
    if multiprocess:
        return _ProcessingType.MULTIPROCESSING
    return _ProcessingType.MULTITHREADING


def parallelize(
    function: Callable[..., _OutputType],
    *params: Iterable[Any],
    workers: Optional[int],
    multiprocess: bool = True,
    **tqdm_kwargs: Any,
) -> List[_OutputType]:
    """Parallelizes the function on given parameters using the specified number of workers.

    :param function: A function to be parallelized.
    :param params: Sequences of parameters to be given to the function. It uses the same logic as Python `map` function.
    :param workers: Maximum number of time the function will be executed in parallel.
    :param multiprocess: If `True` it will use `concurrent.futures.ProcessPoolExecutor` which will distribute
        workflow executions among multiple processors. If `False` it will use
        `concurrent.futures.ThreadPoolExecutor` which will distribute workflow among multiple threads. In case of
        `workers=1` this parameter is ignored and workflows will be executed consecutively.
    :param tqdm_kwargs: Keyword arguments that will be propagated to `tqdm` progress bar.
    :return: A list of function results.
    """
    if not params:
        raise ValueError(
            "At least 1 list of parameters should be given. Otherwise it is not clear how many times the"
            "function has to be executed."
        )
    processing_type = _decide_processing_type(workers=workers, multiprocess=multiprocess)

    if processing_type is _ProcessingType.SINGLE_PROCESS:
        size = len(params[0] if isinstance(params[0], Collection) else list(params[0]))
        return list(tqdm(map(function, *params), total=size, **tqdm_kwargs))

    if processing_type is _ProcessingType.MULTITHREADING:
        with ThreadPoolExecutor(max_workers=workers) as thread_executor:
            return submit_and_monitor_execution(thread_executor, function, *params, **tqdm_kwargs)

    # pylint: disable=global-statement
    global MULTIPROCESSING_LOCK
    try:
        MULTIPROCESSING_LOCK = multiprocessing.Manager().Lock()
        with ProcessPoolExecutor(max_workers=workers) as process_executor:
            return submit_and_monitor_execution(process_executor, function, *params, **tqdm_kwargs)
    finally:
        MULTIPROCESSING_LOCK = None


def execute_with_mp_lock(function: Callable[..., _OutputType], *args: Any, **kwargs: Any) -> _OutputType:
    """A helper utility function that executes a given function with multiprocessing lock if the process is being
    executed in a multiprocessing mode.

    :param function: A function
    :param args: Function's positional arguments
    :param kwargs: Function's keyword arguments
    :return: Function's results
    """
    if multiprocessing.current_process().name == "MainProcess" or MULTIPROCESSING_LOCK is None:
        return function(*args, **kwargs)

    # pylint: disable=not-context-manager
    with MULTIPROCESSING_LOCK:
        return function(*args, **kwargs)


def submit_and_monitor_execution(
    executor: Executor,
    function: Callable[..., _OutputType],
    *params: Iterable[Any],
    **tqdm_kwargs: Any,
) -> List[_OutputType]:
    """Performs the execution parallelization and monitors the process using a progress bar.

    :param executor: An object that performs parallelization.
    :param function: A function to be parallelized.
    :param params: Each element in a sequence are parameters for a single call of `function`.
    :return: A list of results in the same order as input parameters given by `executor_params`.
    """
    futures = [executor.submit(function, *function_params) for function_params in zip(*params)]
    return join_futures(futures, **tqdm_kwargs)


def join_futures(futures: List[Future], **tqdm_kwargs: Any) -> List[Any]:
    """Resolves futures, monitors progress, and returns a list of results.

    :param futures: A list of futures to be joined. Note that this list will be reduced into an empty list as a side
        effect of this function. This way future objects will get cleared from memory already during the execution
        which will free some extra memory. But this can be achieved only if future objects aren't kept in memory
        outside `futures` list.
    :param tqdm_kwargs: Keyword arguments that will be propagated to `tqdm` progress bar.
    :return: A list of results in the order that corresponds with the order of the given input `futures`.
    """
    results: List[Optional[Any]] = [None] * len(futures)
    for position, result in join_futures_iter(futures, **tqdm_kwargs):
        results[position] = result

    return cast(List[Any], results)


def join_futures_iter(
    futures: List[Future], update_interval: float = 0.5, **tqdm_kwargs: Any
) -> Generator[Tuple[int, Any], None, None]:
    """Resolves futures, monitors progress, and serves as an iterator over results.

    :param futures: A list of futures to be joined. Note that this list will be reduced into an empty list as a side
        effect of this function. This way future objects will get cleared from memory already during the execution
        which will free some extra memory. But this can be achieved only if future objects aren't kept in memory
        outside `futures` list.
    :param update_interval: A number of seconds to wait between consecutive updates of a progress bar.
    :param tqdm_kwargs: Keyword arguments that will be propagated to `tqdm` progress bar.
    :return: A generator that will be returning pairs `(index, result)` where `index` will define the position of future
        in the original list to which `result` belongs to.
    """

    def _wait_function(remaining_futures: Collection[Future]) -> Tuple[Collection[Future], Collection[Future]]:
        done, not_done = concurrent.futures.wait(
            remaining_futures, timeout=float(update_interval), return_when=FIRST_COMPLETED
        )
        return done, not_done

    def _get_result(future: Future) -> Any:
        return future.result()

    return _base_join_futures_iter(_wait_function, _get_result, futures, **tqdm_kwargs)


def _base_join_futures_iter(
    wait_function: Callable[[Collection[_FutureType]], Tuple[Collection[_FutureType], Collection[_FutureType]]],
    get_result_function: Callable[[_FutureType], _OutputType],
    futures: List[_FutureType],
    **tqdm_kwargs: Any,
) -> Generator[Tuple[int, _OutputType], None, None]:
    """A generalized utility function that resolves futures, monitors progress, and serves as an iterator over
    results."""
    if not isinstance(futures, list):
        raise ValueError(f"Parameters 'futures' should be a list but {type(futures)} was given")
    remaining_futures: Collection[_FutureType] = _make_copy_and_empty_given(futures)

    id_to_position_map = {id(future): index for index, future in enumerate(remaining_futures)}

    with tqdm(total=len(remaining_futures), **tqdm_kwargs) as pbar:
        while remaining_futures:
            done, remaining_futures = wait_function(remaining_futures)
            for future in done:
                result = get_result_function(future)
                result_position = id_to_position_map[id(future)]
                pbar.update(1)
                yield result_position, result


def _make_copy_and_empty_given(items: List[_T]) -> List[_T]:
    """Removes items from the given list and returns its copy. The side effect of removing items is intentional."""
    items_copy = items[:]
    while items:
        items.pop()
    return items_copy



import os

from setuptools import find_packages, setup


def get_long_description():
    this_directory = os.path.abspath(os.path.dirname(__file__))

    with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
        long_description = f.read()

    return long_description


def parse_requirements(file):
    with open(os.path.join(os.path.dirname(__file__), file)) as requirements_file:
        return sorted({line.partition("#")[0].strip() for line in requirements_file} - set(""))


def get_version():
    path = os.path.join(os.path.dirname(__file__), "eolearn", "core", "__init__.py")
    with open(path) as version_file:
        for line in version_file:
            if line.find("__version__") >= 0:
                version = line.split("=")[1].strip()
                version = version.strip('"').strip("'")
    return version


setup(
    name="eo-learn-core",
    python_requires=">=3.7",
    version=get_version(),
    description="Core Machine Learning Framework at Sinergise",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/sentinel-hub/eo-learn",
    project_urls={
        "Documentation": "https://eo-learn.readthedocs.io",
        "Source Code": "https://github.com/sentinel-hub/eo-learn",
        "Bug Tracker": "https://github.com/sentinel-hub/eo-learn/issues",
        "Forum": "https://forum.sentinel-hub.com",
    },
    author="Sinergise EO research team",
    author_email="eoresearch@sinergise.com",
    license="MIT",
    packages=find_packages(exclude=["eolearn.tests*"]),
    package_data={"eolearn": ["core/py.typed"]},
    include_package_data=True,
    install_requires=parse_requirements("requirements.txt"),
    extras_require={"RAY": parse_requirements("requirements-ray.txt")},
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Scientific/Engineering :: Image Processing",
    ],
)



"""
The following objects and functions are the core of eo-learn package
"""
from .constants import FeatureType, FeatureTypeSet, OverwritePermission
from .core_tasks import (
    AddFeatureTask,
    CopyTask,
    CreateEOPatchTask,
    DeepCopyTask,
    DuplicateFeatureTask,
    ExplodeBandsTask,
    ExtractBandsTask,
    InitializeFeatureTask,
    LoadTask,
    MapFeatureTask,
    MergeEOPatchesTask,
    MergeFeatureTask,
    MoveFeatureTask,
    RemoveFeatureTask,
    RenameFeatureTask,
    SaveTask,
    ZipFeatureTask,
)
from .eodata import EOPatch
from .eoexecution import EOExecutor
from .eonode import EONode, linearly_connect_tasks
from .eotask import EOTask
from .eoworkflow import EOWorkflow, WorkflowResults
from .eoworkflow_tasks import OutputTask
from .utils.common import deep_eq
from .utils.fs import get_filesystem, load_s3_filesystem, pickle_fs, unpickle_fs
from .utils.parallelize import execute_with_mp_lock, join_futures, join_futures_iter, parallelize
from .utils.parsing import FeatureParser

__version__ = "1.3.1"



"""
The utilities module is a collection of classes and functions used across the eolearn package, such as checking whether
two objects are deeply equal, padding of an image, etc.

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Tomislav Slijepčević, Nejc Vesel, Jovan Višnjić (Sinergise)
Copyright (c) 2017-2022 Anže Zupanc (Sinergise)
Copyright (c) 2019-2020 Jernej Puc, Lojze Žust (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

import logging
from logging import Filter, LogRecord
from typing import Any, Optional

LOGGER = logging.getLogger(__name__)


class LogFileFilter(Filter):
    """Filters log messages passed to log file."""

    def __init__(self, thread_name: Optional[str], *args: Any, **kwargs: Any):
        """
        :param thread_name: Name of the thread by which to filter logs. By default, it won't filter by any name.
        """
        self.thread_name = thread_name
        super().__init__(*args, **kwargs)

    def filter(self, record: LogRecord) -> bool:
        """Shows everything from the thread that it was initialized in."""
        return record.threadName == self.thread_name



"""
This module implements feature types used in EOPatch objects

Credits:
Copyright (c) 2017-2022 Matej Aleksandrov, Matej Batič, Grega Milčinski, Domagoj Korais, Matic Lubej (Sinergise)
Copyright (c) 2017-2022 Žiga Lukšič, Devis Peressutti, Tomislav Slijepčević, Nejc Vesel, Jovan Višnjić (Sinergise)
Copyright (c) 2017-2022 Anže Zupanc (Sinergise)
Copyright (c) 2017-2019 Blaž Sovdat, Andrej Burja (Sinergise)

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""
from enum import Enum
from typing import Optional

from sentinelhub import BBox, MimeType


class FeatureType(Enum):
    """The Enum class of all possible feature types that can be included in EOPatch.

    List of feature types:
     - DATA with shape t x n x m x d: time- and position-dependent remote sensing data (e.g. bands) of type float
     - MASK with shape t x n x m x d: time- and position-dependent mask (e.g. ground truth, cloud/shadow mask,
       super pixel identifier) of type int
     - SCALAR with shape t x s: time-dependent and position-independent remote sensing data (e.g. weather data,) of
       type float
     - LABEL with shape t x s: time-dependent and position-independent label (e.g. ground truth) of type int
     - VECTOR: a list of time-dependent vector shapes in shapely.geometry classes
     - DATA_TIMELESS with shape n x m x d: time-independent and position-dependent remote sensing data (e.g.
       elevation model) of type float
     - MASK_TIMELESS with shape n x m x d: time-independent and position-dependent mask (e.g. ground truth,
       region of interest mask) of type int
     - SCALAR_TIMELESS with shape s:  time-independent and position-independent remote sensing data of type float
     - LABEL_TIMELESS with shape s: time-independent and position-independent label of type int
     - VECTOR_TIMELESS: time-independent vector shapes in shapely.geometry classes
     - META_INFO: dictionary of additional info (e.g. resolution, time difference)
     - BBOX: bounding box of the patch which is an instance of sentinelhub.BBox
     - TIMESTAMP: list of dates which are instances of datetime.datetime
    """

    # IMPORTANT: these feature names must exactly match those in EOPatch constructor
    DATA = "data"
    MASK = "mask"
    SCALAR = "scalar"
    LABEL = "label"
    VECTOR = "vector"
    DATA_TIMELESS = "data_timeless"
    MASK_TIMELESS = "mask_timeless"
    SCALAR_TIMELESS = "scalar_timeless"
    LABEL_TIMELESS = "label_timeless"
    VECTOR_TIMELESS = "vector_timeless"
    META_INFO = "meta_info"
    BBOX = "bbox"
    TIMESTAMP = "timestamp"

    @classmethod
    def has_value(cls, value: str) -> bool:
        """True if value is in FeatureType values. False otherwise."""
        return value in cls._value2member_map_

    def is_spatial(self) -> bool:
        """True if FeatureType has a spatial component. False otherwise."""
        return self in FeatureTypeSet.SPATIAL_TYPES

    def is_temporal(self) -> bool:
        """True if FeatureType has a time component. False otherwise."""
        return self in FeatureTypeSet.TEMPORAL_TYPES

    def is_timeless(self) -> bool:
        """True if FeatureType doesn't have a time component and is not a meta feature. False otherwise."""
        return self in FeatureTypeSet.TIMELESS_TYPES

    def is_discrete(self) -> bool:
        """True if FeatureType should have discrete (integer) values. False otherwise."""
        return self in FeatureTypeSet.DISCRETE_TYPES

    def is_meta(self) -> bool:
        """True if FeatureType is for storing metadata info and False otherwise."""
        return self in FeatureTypeSet.META_TYPES

    def is_vector(self) -> bool:
        """True if FeatureType is vector feature type. False otherwise."""
        return self in FeatureTypeSet.VECTOR_TYPES

    def has_dict(self) -> bool:
        """True if FeatureType stores a dictionary. False otherwise."""
        return self in FeatureTypeSet.DICT_TYPES

    def is_raster(self) -> bool:
        """True if FeatureType stores a dictionary with raster data. False otherwise."""
        return self in FeatureTypeSet.RASTER_TYPES

    def contains_ndarrays(self) -> bool:
        """True if FeatureType stores a dictionary of numpy.ndarrays. False otherwise."""
        return self in FeatureTypeSet.RASTER_TYPES

    def ndim(self) -> Optional[int]:
        """If given FeatureType stores a dictionary of numpy.ndarrays it returns dimensions of such arrays."""
        if self.is_raster():
            return {
                FeatureType.DATA: 4,
                FeatureType.MASK: 4,
                FeatureType.SCALAR: 2,
                FeatureType.LABEL: 2,
                FeatureType.DATA_TIMELESS: 3,
                FeatureType.MASK_TIMELESS: 3,
                FeatureType.SCALAR_TIMELESS: 1,
                FeatureType.LABEL_TIMELESS: 1,
            }[self]
        return None

    def type(self) -> type:
        """Returns type of the data for the given FeatureType."""
        if self is FeatureType.TIMESTAMP:
            return list
        if self is FeatureType.BBOX:
            return BBox
        return dict

    def file_format(self) -> MimeType:
        """Returns a mime type enum of a file format into which data of the feature type will be serialized"""
        if self.is_raster():
            return MimeType.NPY
        if self.is_vector():
            return MimeType.GPKG
        if self is FeatureType.BBOX:
            return MimeType.GEOJSON
        return MimeType.JSON


class FeatureTypeSet:
    """A collection of immutable sets of feature types, grouped together by certain properties."""

    SPATIAL_TYPES = frozenset(
        [
            FeatureType.DATA,
            FeatureType.MASK,
            FeatureType.VECTOR,
            FeatureType.DATA_TIMELESS,
            FeatureType.MASK_TIMELESS,
            FeatureType.VECTOR_TIMELESS,
        ]
    )
    TEMPORAL_TYPES = frozenset(
        [
            FeatureType.DATA,
            FeatureType.MASK,
            FeatureType.SCALAR,
            FeatureType.LABEL,
            FeatureType.VECTOR,
            FeatureType.TIMESTAMP,
        ]
    )
    TIMELESS_TYPES = frozenset(
        [
            FeatureType.DATA_TIMELESS,
            FeatureType.MASK_TIMELESS,
            FeatureType.SCALAR_TIMELESS,
            FeatureType.LABEL_TIMELESS,
            FeatureType.VECTOR_TIMELESS,
        ]
    )
    DISCRETE_TYPES = frozenset(
        [FeatureType.MASK, FeatureType.MASK_TIMELESS, FeatureType.LABEL, FeatureType.LABEL_TIMELESS]
    )
    META_TYPES = frozenset([FeatureType.META_INFO, FeatureType.BBOX, FeatureType.TIMESTAMP])
    VECTOR_TYPES = frozenset([FeatureType.VECTOR, FeatureType.VECTOR_TIMELESS])
    RASTER_TYPES = frozenset(
        [
            FeatureType.DATA,
            FeatureType.MASK,
            FeatureType.SCALAR,
            FeatureType.LABEL,
            FeatureType.DATA_TIMELESS,
            FeatureType.MASK_TIMELESS,
            FeatureType.SCALAR_TIMELESS,
            FeatureType.LABEL_TIMELESS,
        ]
    )
    DICT_TYPES = frozenset(
        [
            FeatureType.DATA,
            FeatureType.MASK,
            FeatureType.SCALAR,
            FeatureType.LABEL,
            FeatureType.VECTOR,
            FeatureType.DATA_TIMELESS,
            FeatureType.MASK_TIMELESS,
            FeatureType.SCALAR_TIMELESS,
            FeatureType.LABEL_TIMELESS,
            FeatureType.VECTOR_TIMELESS,
            FeatureType.META_INFO,
        ]
    )
    RASTER_TYPES_4D = frozenset([FeatureType.DATA, FeatureType.MASK])
    RASTER_TYPES_3D = frozenset([FeatureType.DATA_TIMELESS, FeatureType.MASK_TIMELESS])
    RASTER_TYPES_2D = frozenset([FeatureType.SCALAR, FeatureType.LABEL])
    RASTER_TYPES_1D = frozenset([FeatureType.SCALAR_TIMELESS, FeatureType.LABEL_TIMELESS])


class OverwritePermission(Enum):
    """Enum class which specifies which content of saved EOPatch can be overwritten when saving new content.

    Permissions are in the following hierarchy:

    - `ADD_ONLY` - Only new features can be added, anything that is already saved cannot be changed.
    - `OVERWRITE_FEATURES` - Overwrite only data for features which have to be saved. The remaining content of saved
      EOPatch will stay unchanged.
    - `OVERWRITE_PATCH` - Overwrite entire content of saved EOPatch and replace it with the new content.
    """

    ADD_ONLY = 0
    OVERWRITE_FEATURES = 1
    OVERWRITE_PATCH = 2
